/**
 * Except where noted, most of the following is drawn from `conio4gcc` by
 * @pushkar6723 <https://github.com/pushkar147/conio4gcc>
 * @see LICENSE.txt
 */
#include <stdio.h>
#include <stdlib.h>
#include "xconio.h"

#if !defined(_WIN32) && !defined(__WIN32__)
void clrscr(void)
{
    system("tput reset");
}

/**
 * @see https://github.com/pushkar147/conio4gcc
 */
void gotoxy(short x, short y)
{
    printf("%c[%hd;%hdf", 0x1B, y, x);
}

/**
 * @see https://github.com/pushkar147/conio4gcc
 */
int getch(void)
{
    struct termios oldattr, newattr;
    int ch;

    tcgetattr(STDIN_FILENO, &oldattr);
    newattr = oldattr;
    newattr.c_lflag &= ~(ICANON | ECHO);
    tcsetattr(STDIN_FILENO, TCSANOW, &newattr);
    ch = getchar();
    tcsetattr(STDIN_FILENO, TCSANOW, &oldattr);

    return ch;
}

/** Adapted from a code example posted to ResearchGate by Konrad Burnik:
 * https://www.researchgate.net/post/How_to_use_arrow_keys_in_c_programming_language
 */
short getkey(void)
{
    short keyvalue, i;

    for (i = 0; i < 3; i++) {
        keyvalue = (short)getch();
        if (i == 0 && ( keyvalue == 0x72 || keyvalue == 0x52 ||\
                        keyvalue == 0x73 || keyvalue == 0x53 ||\
                        keyvalue == 0x71 || keyvalue == 0x51) ) {
            break;
        } else if (i == 0 && keyvalue != 0x1B)
            break;
    }

    return keyvalue;
}
#else

void clrscr(void)
{
    system("cls");
}

/**
 * @copyright (c) 2011 Benjamin Lindley
 * Distributed under the terms of the Creative Commons Attribution-ShareAlike
 * 3.0 Unported License
 * @see https://stackoverflow.com/a/7955343
 */
void gotoxy(short x, short y)
{
    static HANDLE h = NULL;
    COORD c         = { x, y };
    if (!h)
        h = GetStdHandle(STD_OUTPUT_HANDLE);

    SetConsoleCursorPosition(h, c);
}

short getkey(void)
{
    short keyvalue, keypress1;
    keypress1 = (short)getch();

    switch(keypress1) {
    case 0xE0:
        keyvalue = (short)getch();
        break;
    case 0x71:
    case 0x51:
    case 0x72:
    case 0x52:
    case 0x73:
    case 0x53:
        keyvalue = keypress1;
        break;
    default:
        keyvalue = 0;
        break;
    }

    return keyvalue;
}
#endif

/**
 * @copyright (c) 1997-2020 QPBC
 * @see http://www.c-for-dummies.com/caio/fpurge.php
 */
void flushnl(void)
{
    int input;
    do
        input = getchar();
    while (input != EOF && input != '\n');
    clearerr(stdin);
}
