#include "puzzle.h"

static unsigned short r, new_row, down_shift, backspaces, moves,
    L_moves, /* needed to accurately track the position of the NULL separator
              * whenever the player moves left, then doubles back to the right;
              * it also keeps the counter from missing the pos. of NULL when
              * moving moving up, then left, from a middle column (i.e. an
              * 'L_moves' value of 1 or 2); see below, esp. lines 33ff.*/

    bump,      /* signals the need to correct an observed shortfall in the
                * pos. of NULL separator if player bumps against the top border and
                * then moves down */
    recovered; /* signals that an edge of the board was hit on a previous move
                  */

static inline void warn(const char *msg)
{
  gotoxy(14, 29);
  scrSetColors(scrRed, scrBlack);
  printf("%s", msg);
}

Choice Play(unsigned short *pNumbers)
{
  short temp = 0, temp1, temp2;
  Choice replay = REPLAY;
  short choice = getkey();

  switch (choice) {
#if defined(_WIN32)
  case 75: /* Win/DOS value of 2nd char sent to buffer by LEFT arrow key */
#else
  case 68: /* Unix value of 3rd char sent to buffer by LEFT arrow key */
#endif
    if (L_moves < 2) /* i.e. only if player has not reached left edge */
      L_moves++;

    if (moves <= 4) {
      moves++;
      if (new_row) { /* if 'new_row' becomes true while the active tile is in a
                      * middle column (i.e. 'L_moves' == 1|2) and the row is
                      * higher than the base (r > 0), the value of 'moves' will
                      * be too small by one; the complex conditon at line 165
                      * tries to account for this. */
        moves--;
        new_row = false;
      }
    }

    if (*(pNumbers + CUR_POS(r) - moves) == *pNumbers ||
        *(pNumbers + ADJ_POS(r) - moves) == *(pNumbers + 11) ||
        *(pNumbers + ADJ_POS(r) - moves) == *(pNumbers + 7) ||
        *(pNumbers + ADJ_POS(r) - moves) == *(pNumbers + 3)) {
      /*i.e., if NULL has reached the right-hand edge of the table */
      warn("Can't move there! \b\a"); /* leave a space for the cursor */

      if (moves <= 4)
        moves--;

      /* allowing 'moves' to reach 4 is crucial for function of 'backspaces' */
      if (L_moves > 0 && L_moves > 2)
        L_moves--;
    } else {
      temp1 = *(pNumbers + CUR_POS(r) - moves) = 0; /* copy NULL separator */
      temp2 = *(pNumbers + ADJ_POS(r) - moves);     /* copy # to left of NULL */
      *(pNumbers + CUR_POS(r) - moves) = temp2;     /* switch 'em */
      *(pNumbers + ADJ_POS(r) - moves) = temp1;
      Board(pNumbers);
      down_shift = false;
    }
    break;
#if defined(_WIN32)
  case 77: /*move number RIGHT */
#else
  case 67:
#endif
    if ((moves < 4 && !backspaces) || (moves <= 4 && down_shift)) {
      /* the second term of this condition accounts for a specific scenario:
       * moving down three times from the top row (allowing 'moves' to reach
       * 4), then right */
      moves++;

      if (new_row && L_moves > 0) {
        /* this condition anticipates moving up as the first move of the game
         * (when 'L_moves' is 0), then trying to go right, then moving up again:
         * the counter will be diminished once here, and a second time by the
         * bounds-checking conditions below, making the counter too short by 1
         */
        moves--;
        new_row = false;
      } else if (new_row)
        moves--;
      /* notice that the second option allows 'new_row' to remain true, so that
       * the value of 'moves' will hold steady if the player hits the right edge
       */
    } else {
      if (*(pNumbers + CUR_POS(r) - moves) != *(pNumbers + 8) &&
          *(pNumbers + CUR_POS(r) - moves) != *(pNumbers + 4) &&
          *(pNumbers + CUR_POS(r) - moves) != *pNumbers) {
        /* this condition anticipate moving left into the first column
         * (elements 8, 4, and 0), then up (i.e.'moves' >= 4 and 'new_row' ==
         * true), in which case the  value of 'moves' is already where it
         * should be; the following prepares for a second move to the right
         * from the pos. just described */
        moves--;
        backspaces++;
      }
    }

    if (backspaces > 2) /* i.e. if player makes two consecutive right moves from
                         * left edge of board */
      backspaces = 0;

    if (*(pNumbers + CUR_POS(r) - moves) == *(pNumbers + 15) ||
        *(pNumbers + CUR_POS(r) - moves) == *(pNumbers + 11) ||
        *(pNumbers + CUR_POS(r) - moves) == *(pNumbers + 7) ||
        *(pNumbers + CUR_POS(r) - moves) == *(pNumbers + 3)) {
      /*i.e., if NULL has reached the left-hand edge of the table */
      warn("Can't move there! \b\a"); /* leave a space for the cursor */

      if (moves > 0 && !new_row)
        moves--;
    } else {
      temp1 = *(pNumbers + CUR_POS(r) - moves) = 0;   /* copy NULL */
      temp2 = *(pNumbers + (CUR_POS(r) - moves) + 1); /*copy # to the */
      *(pNumbers + CUR_POS(r) - moves) = temp2;       /* switch */
      *(pNumbers + (CUR_POS(r) - moves) + 1) = temp1;
      down_shift = false;
      Board(pNumbers);

      if (!new_row && L_moves > 0) /* i.e. make sure counter reaches edge of
                                      board even if the player doubles back
                                      after moving one, two or three tiles to
                                      the left */
        moves -= 2;

      if (new_row) { /* i.e. if player has previously moved into the first
                      * column from the left, then up, then right, then moves
                      * right a second time, so that 'moves' is now >= 4 and
                      * 'new_row' still tests true, having skipped over the
                      * nested 'if' and 'else-if' conditions at lines 80 and 88
                      * above */
        moves -= 2;
        new_row = false;
      }
    }
    break;
#if defined(_WIN32)
  case 72: /* . . . UP */
#else
  case 65:
#endif
    if (r == 3) {                     /* i.e., if active tile is in top row */
      warn("Can't move there! \b\a"); /* leave a space for the cursor */
      bump = true;
    } else {
      if (new_row) { /* i.e., if player has just moved up and decides to move up
                      * again, in which case the 'new_row' var will still be
                      * true, since moving left or right always resets 'new_row'
                      * to false */
        temp = L_moves;
        L_moves = 0;
      }

      /* simply voiding the value of 'L_moves' will cause problems later on;
       * we just want to prevent incrementation of 'moves' in the above
       * scenario */
      if (moves < 4 && ((r == 0 && !down_shift) || (r > 0 && L_moves > 0))) {
        moves++;
        new_row = true;
      }

      temp1 = *(pNumbers + CUR_POS(r) - moves) = 0;   /*copy NULL */
      temp2 = *(pNumbers + (CUR_POS(r) - moves) - 4); /*copy upper adjacent */
      *(pNumbers + CUR_POS(r) - moves) = temp2;       /* pull number down */
      *(pNumbers + (CUR_POS(r) - moves) - 4) = temp1; /* fill vacancy */
      Board(pNumbers);
      r++;

      if (temp)
        L_moves = temp; /* restore value, if any, of 'L_moves' */
    }
    break;
#if defined(_WIN32)
  case 80: /* DOWN */
#else
  case 66:
#endif
    if (r == 0) /* i.e., if active tile is in bottom row */
      warn("Can't move there! \b\a");
    else {
      if (r > 0 && !down_shift && !new_row && moves <= 4)
        moves++;

      if (r == 3) {
        if (bump) {
          if (L_moves == 0 || down_shift || (L_moves > 0 && !down_shift))
            moves--;
          else
            moves++;

          recovered = true;
          bump = false;
        }

        if (recovered) { /* i.e., moving left or right after hitting top border
                            */
          moves++;
          recovered = false;
        }
      }

      down_shift = true;
      new_row = true;
      temp1 = *(pNumbers + CUR_POS(r) - moves) = 0;   /*copy NULL */
      temp2 = *(pNumbers + (CUR_POS(r) - moves) + 4); /*copy lower adjacent # */
      *(pNumbers + CUR_POS(r) - moves) = temp2;       /* pull # up */
      *(pNumbers + (CUR_POS(r) - moves) + 4) = temp1; /* fill vacancy */
      r--;
      Board(pNumbers);
    }
    break;
  case 114:
  case 82:
    moves = 0;
    L_moves = 0;
    backspaces = 0;
    r = 0;
    new_row = false;
    down_shift = false;
    bump = false;
    replay = NEW_GAME;
    break;
  case 115:
  case 83:
    if (Solve()) {
      moves = 0;
      L_moves = 0;
      down_shift = 0;
      backspaces = 0;
      r = 0;
      new_row = 0;
      bump = 0;
      replay = NEW_GAME;
    } else
      replay = QUIT;
    break;
  case 113:
  case 81:
    gotoxy(10 + MARGIN, 8);
    scrSetColors(scrGreen, scrBlack);
    printf("Quitting. . . ");
#if !defined(_WIN32) && defined(__GNUC__)
    fflush(stdout);
#endif
    replay = QUIT;
    break;
  default:
    warn("?");
    scrSetColors(scrWhite, scrBlack);
  }

  return (replay);
}
